"""
Tests for the ApplicationProperties class
"""
from pymarkdown.application_properties import (
    ApplicationProperties,
    ApplicationPropertiesFacade,
)

# pylint: disable=too-many-lines


def test_property_name_separator():
    """
    Test to make sure that the property name separator is as expected.
    """

    # Arrange
    application_properties = ApplicationProperties()
    expected_separator = "."

    # Act
    actual_separator = application_properties.separator

    # Assert
    assert actual_separator == expected_separator


def test_properties_with_object():
    """
    Test to make sure that a default application properties object has no properties.
    """

    # Arrange
    application_properties = ApplicationProperties()
    expected_property_count = 0

    # Act
    actual_property_count = application_properties.number_of_properties

    # Assert
    assert actual_property_count == expected_property_count


def test_properties_with_single_property():
    """
    Test a configuration map with a single property, and how that property looks.
    """

    # Arrange
    application_properties = ApplicationProperties()
    config_map = {"enabled": True}
    expected_property_count = 1

    # Act
    application_properties.load_from_dict(config_map)
    actual_property_count = application_properties.number_of_properties
    found_names = application_properties.property_names

    # Assert
    assert actual_property_count == expected_property_count
    assert len(found_names) == expected_property_count
    assert "enabled" in found_names


def test_properties_with_single_nested_property():
    """
    Test a configuration map with a single nested property, and how that property looks.
    """

    # Arrange
    application_properties = ApplicationProperties()
    config_map = {"feature": {"enabled": True}}
    expected_property_count = 1

    # Act
    application_properties.load_from_dict(config_map)
    actual_property_count = application_properties.number_of_properties
    found_names = application_properties.property_names

    # Assert
    assert actual_property_count == expected_property_count
    assert len(found_names) == expected_property_count
    assert "feature.enabled" in found_names


def test_properties_with_mixed_properties():
    """
    Test a configuration map with properties at different levels, and how those properties look.
    """

    # Arrange
    application_properties = ApplicationProperties()
    config_map = {
        "feature": {"enabled": True},
        "other_feature": {"enabled": False, "other": 1},
    }
    expected_property_count = 3

    # Act
    application_properties.load_from_dict(config_map)
    actual_property_count = application_properties.number_of_properties
    found_names = application_properties.property_names

    # Assert
    assert actual_property_count == expected_property_count
    assert len(found_names) == expected_property_count
    assert "feature.enabled" in found_names
    assert "other_feature.enabled" in found_names
    assert "other_feature.other" in found_names


def test_properties_load_from_non_dictionary():
    """
    Test a loading a configuration map that is not a dictionary.
    """

    # Arrange
    application_properties = ApplicationProperties()
    config_map = [{"feature": True}]

    # Act
    raised_exception = None
    try:
        application_properties.load_from_dict(config_map)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception) == "Specified parameter was not a dictionary."
    ), "Expected message was not present in exception."


def test_properties_load_with_non_string_key():
    """
    Test a loading a configuration map that contains a key that is not a string.
    """

    # Arrange
    application_properties = ApplicationProperties()
    config_map = {1: True}

    # Act
    raised_exception = None
    try:
        application_properties.load_from_dict(config_map)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "All keys in the main dictionary and nested dictionaries must be strings."
    ), "Expected message was not present in exception."


def test_properties_load_with_key_containing_dot():
    """
    Test a loading a configuration map that contains a key with a '.' character.
    """

    # Arrange
    application_properties = ApplicationProperties()
    config_map = {"my.property": True}

    # Act
    raised_exception = None
    try:
        application_properties.load_from_dict(config_map)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Keys strings cannot contain the separator character '.'."
    ), "Expected message was not present in exception."


def test_properties_get_generic_with_bad_type():
    """
    Test a fetching a configuration value where the generic function is
    used and the type and the default are confused.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_property("property", False)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The property_type argument for 'property' must be a type."
    ), "Expected message was not present in exception."


def test_properties_get_generic_with_required_and_found():
    """
    Test a fetching a configuration value where the value is required and present.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = True

    # Act
    actual_value = application_properties.get_property(
        "property", bool, is_required=True
    )

    # Assert
    assert actual_value == expected_value


def test_properties_get_generic_with_required_and_not_found():
    """
    Test a fetching a configuration value where the value is required and not present.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_property("other_property", bool, is_required=True)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "A value for property 'other_property' must be provided."
    ), "Expected message was not present in exception."


def test_properties_get_generic_with_strict_mode_and_bad_type():
    """
    Test a fetching a configuration value where strict mode is on and the type is not correct.
    """

    # Arrange
    config_map = {"property": 1}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_property("property", str, strict_mode=True)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The value for property 'property' must be of type 'str'."
    ), "Expected message was not present in exception."


def test_properties_get_generic_with_global_strict_mode_and_bad_type():
    """
    Test a fetching a configuration value where strict mode is on and the type is not correct.
    """

    # Arrange
    config_map = {"property": 1}
    application_properties = ApplicationProperties(strict_mode=True)
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_property("property", str)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert application_properties.strict_mode
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The value for property 'property' must be of type 'str'."
    ), "Expected message was not present in exception."


def sample_string_validation_function(property_value):
    """
    Simple string validation that throws an error if not "1" or "2".
    """
    if property_value not in ["1", "2"]:
        raise ValueError("Value '" + str(property_value) + "' is not '1' or '2'")


def test_properties_get_generic_with_strict_mode_and_bad_validity():
    """
    Test a fetching a configuration value where strict mode is on and the value is not valid.
    """

    # Arrange
    config_map = {"property": "3"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_property(
            "property",
            str,
            strict_mode=True,
            valid_value_fn=sample_string_validation_function,
        )
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The value for property 'property' is not valid: Value '3' is not '1' or '2'"
    ), "Expected message was not present in exception."


def test_properties_get_boolean_with_found_value():
    """
    Test fetching a configuration value that is present and boolean.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = True

    # Act
    actual_value = application_properties.get_boolean_property("property", False)

    # Assert
    assert expected_value == actual_value


def test_properties_get_boolean_with_found_value_but_wrong_type():
    """
    Test fetching a configuration value that is present and not boolean.
    """

    # Arrange
    config_map = {"property": 1}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = True

    # Act
    actual_value = application_properties.get_boolean_property("property", True)

    # Assert
    assert expected_value == actual_value


def test_properties_get_boolean_with_not_found_value():
    """
    Test fetching a configuration value that is not present and boolean.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = True

    # Act
    actual_value = application_properties.get_boolean_property("other_property", True)

    # Assert
    assert expected_value == actual_value


def test_properties_get_boolean_with_not_found_value_and_no_default_value():
    """
    Test fetching a configuration value that is not present, with no default, and boolean.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    actual_value = application_properties.get_boolean_property("other_property")

    # Assert
    assert actual_value is None


def test_properties_get_boolean_with_a_bad_property_name():
    """
    Test fetching a configuration value with a bad property name.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_boolean_property(1, 1)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception) == "The propertyName argument must be a string."
    ), "Expected message was not present in exception."


def test_properties_get_boolean_with_a_bad_default():
    """
    Test fetching a configuration value with a default value that is not a boolean.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_boolean_property("property", 1)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The default value for property 'property' must either be None or a 'bool' value."
    ), "Expected message was not present in exception."


def test_properties_get_integer_with_found_value():
    """
    Test fetching a configuration value that is present and integer.
    """

    # Arrange
    config_map = {"property": 1}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = 1

    # Act
    actual_value = application_properties.get_integer_property("property", -1)

    # Assert
    assert expected_value == actual_value


def test_properties_get_integer_with_found_value_but_wrong_type():
    """
    Test fetching a configuration value that is present and not integer.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = -1

    # Act
    actual_value = application_properties.get_integer_property("property", -1)

    # Assert
    assert expected_value == actual_value


def test_properties_get_integer_with_not_found_value():
    """
    Test fetching a configuration value that is not present and integer.
    """

    # Arrange
    config_map = {"property": 2}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = 3

    # Act
    actual_value = application_properties.get_integer_property("other_property", 3)

    # Assert
    assert expected_value == actual_value


def test_properties_get_integer_with_not_found_value_and_no_default_value():
    """
    Test fetching a configuration value that is not present, with no default, and integer.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    actual_value = application_properties.get_integer_property("other_property")

    # Assert
    assert actual_value is None


def test_properties_get_integer_with_a_bad_property_name():
    """
    Test fetching a configuration value with a bad property name.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_integer_property(1, 1)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception) == "The propertyName argument must be a string."
    ), "Expected message was not present in exception."


def test_properties_get_integer_with_a_bad_default():
    """
    Test fetching a configuration value with a default value that is not an integer.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_integer_property("property", True)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The default value for property 'property' must either be None or a 'int' value."
    ), "Expected message was not present in exception."


def test_properties_get_string_with_found_value():
    """
    Test fetching a configuration value that is present and string.
    """

    # Arrange
    config_map = {"property": "me"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = "me"

    # Act
    actual_value = application_properties.get_string_property("property", "")

    # Assert
    assert expected_value == actual_value


def test_properties_get_string_with_found_value_but_wrong_type():
    """
    Test fetching a configuration value that is present and not string.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = ""

    # Act
    actual_value = application_properties.get_string_property("property", "")

    # Assert
    assert expected_value == actual_value


def test_properties_get_string_with_not_found_value():
    """
    Test fetching a configuration value that is not present and string.
    """

    # Arrange
    config_map = {"property": "2"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = "3"

    # Act
    actual_value = application_properties.get_string_property("other_property", "3")

    # Assert
    assert expected_value == actual_value


def test_properties_get_string_with_not_found_value_and_no_default_value():
    """
    Test fetching a configuration value that is not present, with no default, and integer.
    """

    # Arrange
    config_map = {"property": "2"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    actual_value = application_properties.get_string_property("other_property")

    # Assert
    assert actual_value is None


def test_properties_get_string_with_found_value_validated():
    """
    Test fetching a configuration value that is present and adheres to the validation function.
    """

    # Arrange
    config_map = {"property": "2"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = "2"

    # Act
    actual_value = application_properties.get_string_property(
        "property", "-", lambda property_value: property_value in ["1", "2"]
    )

    # Assert
    assert expected_value == actual_value


def test_properties_get_string_with_found_value_not_validated():
    """
    Test fetching a configuration value that is present and does not adhere to the validation function.
    """

    # Arrange
    config_map = {"property": "3"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = "-"

    # Act
    actual_value = application_properties.get_string_property(
        "property", "-", sample_string_validation_function
    )

    # Assert
    assert expected_value == actual_value


def bad_validation_function(property_value):
    """
    Test validation function that always throws an exception.
    """
    raise Exception("huh? " + str(property_value))


def test_properties_get_string_with_found_value_validation_raises_error():
    """
    Test fetching a configuration value that is present and the validation function raises an error.
    """

    # Arrange
    config_map = {"property": "1"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    expected_value = "-"

    # Act
    actual_value = application_properties.get_string_property(
        "property", "-", bad_validation_function
    )

    # Assert
    assert expected_value == actual_value


def test_properties_get_string_with_a_bad_property_name():
    """
    Test fetching a configuration value with a bad property name.
    """

    # Arrange
    config_map = {"property": True}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_string_property(1, "3")
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception) == "The propertyName argument must be a string."
    ), "Expected message was not present in exception."


def test_properties_get_string_with_a_bad_default():
    """
    Test fetching a configuration value with a default value that is not a string.
    """

    # Arrange
    config_map = {"property": "2"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        application_properties.get_string_property("property", True)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The default value for property 'property' must either be None or a 'str' value."
    ), "Expected message was not present in exception."


def test_properties_facade_base_not_properties_object():
    """
    Test setting up a facade with properties object that is not a properties object.
    """

    # Arrange

    # Act
    raised_exception = None
    try:
        ApplicationPropertiesFacade(1, 1)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The base_properties of the facade must be an ApplicationProperties instance."
    ), "Expected message was not present in exception."


def test_properties_facade_prefix_not_string():
    """
    Test setting up a facade with a prefix that is not a string.
    """

    # Arrange
    config_map = {"property": "2"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        ApplicationPropertiesFacade(application_properties, 1)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception) == "The property_prefix argument must be a string."
    ), "Expected message was not present in exception."


def test_properties_facade_prefix_not_terminated_with_separator():
    """
    Test setting up a facade with a prefix that is not terminated with the separator.
    """

    # Arrange
    config_map = {"property": "2"}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)

    # Act
    raised_exception = None
    try:
        ApplicationPropertiesFacade(application_properties, "my")
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "The property_prefix argument must end with the separator character '.'."
    ), "Expected message was not present in exception."


def test_properties_facade_get_with_found_value():
    """
    Test fetching through a configuration facade for a property that is present.
    """

    # Arrange
    config_map = {"upper": {"property": 1.2}}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    facade = ApplicationPropertiesFacade(application_properties, "upper.")
    expected_value = 1.2

    # Act
    actual_value = facade.get_property("property", float)

    # Assert
    assert expected_value == actual_value


def test_properties_facade_get_boolean_with_found_value():
    """
    Test fetching through a configuration facade for a boolean property that is present.
    """

    # Arrange
    config_map = {"upper": {"property": True}}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    facade = ApplicationPropertiesFacade(application_properties, "upper.")
    expected_value = True

    # Act
    actual_value = facade.get_boolean_property("property")

    # Assert
    assert expected_value == actual_value


def test_properties_facade_get_integer_with_found_value():
    """
    Test fetching through a configuration facade for an integer property that is present.
    """

    # Arrange
    config_map = {"upper": {"property": 2}}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    facade = ApplicationPropertiesFacade(application_properties, "upper.")
    expected_value = 2

    # Act
    actual_value = facade.get_integer_property("property")

    # Assert
    assert expected_value == actual_value


def test_properties_facade_get_string_with_found_value():
    """
    Test fetching through a configuration facade for a string property that is present.
    """

    # Arrange
    config_map = {"upper": {"property": "2"}}
    application_properties = ApplicationProperties()
    application_properties.load_from_dict(config_map)
    facade = ApplicationPropertiesFacade(application_properties, "upper.")
    expected_value = "2"

    # Act
    actual_value = facade.get_string_property("property")

    # Assert
    assert expected_value == actual_value


def test_properties_set_manual_property_with_non_string():
    """
    Test to make sure that a manual property with a non-string is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_string = 1

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Manual property form must either be a string or an iterable of strings."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_no_equals():
    """
    Test to make sure that a full key with no value part is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_string = "a_property"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Manual property key and value must be separated by the '=' character."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_single_part_string():
    """
    Test to make sure that a full key specifying an string without any format is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = "123"
    full_string = f"{full_property_key}={property_value}"

    # Act
    application_properties.set_manual_property(full_string)

    # Assert
    actual_value = application_properties.get_string_property(full_property_key)
    assert property_value == actual_value


def test_properties_set_manual_property_with_only_format_indicator():
    """
    Test to make sure that a full key specifying an string format but with no format character following is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = "$123"
    full_string = f"{full_property_key}={property_value}"

    # Act
    application_properties.set_manual_property(full_string)

    # Assert
    actual_value = application_properties.get_string_property(full_property_key)
    assert property_value[1:] == actual_value


def test_properties_set_manual_property_with_string_indicator():
    """
    Test to make sure that a full key specifying an string is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = "$$123"
    full_string = f"{full_property_key}={property_value}"

    # Act
    application_properties.set_manual_property(full_string)

    # Assert
    actual_value = application_properties.get_string_property(full_property_key)
    assert property_value[2:] == actual_value


def test_properties_set_manual_property_with_integer_indicator():
    """
    Test to make sure that a full key specifying an integer is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}"

    # Act
    application_properties.set_manual_property(full_string)

    # Assert
    actual_value = application_properties.get_integer_property(full_property_key)
    assert property_value == actual_value


def test_properties_set_manual_property_with_integer_indicator_and_bad_integer():
    """
    Test to make sure that a full key specifying a bad integer is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}a"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Manual property value '$#123a' cannot be translated into an integer."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_boolean_indicator():
    """
    Test to make sure that a full key specifying a True boolean is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = True
    full_string = f"{full_property_key}=$!{property_value}"

    # Act
    application_properties.set_manual_property(full_string)

    # Assert
    actual_value = application_properties.get_boolean_property(full_property_key)
    assert property_value == actual_value


def test_properties_set_manual_property_with_uncased_boolean_indicator():
    """
    Test to make sure that a full key specifying a True boolean in lower case is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = True
    full_string = f"{full_property_key}=$!{str(property_value).lower()}"

    # Act
    application_properties.set_manual_property(full_string)

    # Assert
    actual_value = application_properties.get_boolean_property(full_property_key)
    assert property_value == actual_value


def test_properties_set_manual_property_with_multiples():
    """
    Test to make sure that a list of full keys is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a"
    property_value = True
    full_string = f"{full_property_key}=$!{property_value}"

    # Act
    application_properties.set_manual_property([full_string])

    # Assert
    actual_value = application_properties.get_boolean_property(full_property_key)
    assert property_value == actual_value


def test_properties_set_manual_property_with_bad_key_start():
    """
    Test to make sure that a full key starting with the key separator character is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = ".a"
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Full property key must not start or end with the '.' character."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_bad_key_end():
    """
    Test to make sure that a full key ending with the key separator character is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a."
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Full property key must not start or end with the '.' character."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_empty_key_middle():
    """
    Test to make sure that an empty key part for the full key is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a..a"
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Full property key cannot contain multiples of the . without any text between them."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_empty_key():
    """
    Test to make sure that a key with an empty key part is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = ""
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Each part of the property key must contain at least one character."
    ), "Expected message was not present in exception."


def test_properties_set_manual_property_with_whitespace_key():
    """
    Test to make sure that a key with whitespace is handled properly.
    """

    # Arrange
    application_properties = ApplicationProperties()
    full_property_key = "a a"
    property_value = 123
    full_string = f"{full_property_key}=$#{property_value}"

    # Act
    raised_exception = None
    try:
        application_properties.set_manual_property(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception)
        == "Each part of the property key must not contain a whitespace character or the '.' character."
    ), "Expected message was not present in exception."


def test_properties_verify_manual_property_form_with_non_string():
    """
    Test to make sure that if we try and test the verification form function
    with a non-string, it fails predictably.
    """

    # Arrange
    full_string = 1

    # Act
    raised_exception = None
    try:
        ApplicationProperties.verify_manual_property_form(full_string)
        assert False, "Should have raised an exception by now."
    except ValueError as this_exception:
        raised_exception = this_exception

    # Assert
    assert raised_exception, "Expected exception was not raised."
    assert (
        str(raised_exception) == "Manual property form must be a string."
    ), "Expected message was not present in exception."



"""
Module to provide tests related to the "-l" option.
"""
from test.markdown_scanner import MarkdownScanner


def test_markdown_with_dash_h():
    """
    Test to make sure we get help if '-h' is supplied
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-h"]

    expected_return_code = 0
    expected_output = """usage: main.py scan [-h] [-l] [-r] path [path ...]

positional arguments:
  path              one or more paths to scan for eligible Markdown files

optional arguments:
  -h, --help        show this help message and exit
  -l, --list-files  list the markdown files found and exit
  -r, --recurse     recursively scan directories
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_only():
    """
    Test to make sure we get help if '-l' is supplied without any paths
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l"]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] path [path ...]
main.py scan: error: the following arguments are required: path
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_bad_path():
    """
    Test to make sure we get help if '-l' is supplied with a bad path.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "my-bad-path"]

    expected_return_code = 1
    expected_output = ""
    expected_error = """Provided path 'my-bad-path' does not exist.
"""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_non_md_directory():
    """
    Test to make sure we get help if '-l' is supplied with a path containing no md files.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "only-text"]

    expected_return_code = 1
    expected_output = ""
    expected_error = """No matching files found.
"""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_md_directory():
    """
    Test to make sure we get help if '-l' is supplied with a path containing a simple md file.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "simple"]

    expected_return_code = 0
    expected_output = """simple/simple.md
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_mixed_directories():
    """
    Test to make sure we get help if '-l' is supplied with a path containing the md directory and the non-md directory.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "only-text", "simple"]

    expected_return_code = 0
    expected_output = """simple/simple.md
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_non_md_file():
    """
    Test to make sure we get help if '-l' is supplied with a file path that isn't a md file.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "only-text/simple_text_file.txt"]

    expected_return_code = 1
    expected_output = ""
    expected_error = """Provided file path 'only-text/simple_text_file.txt' is not a valid file. Skipping.
"""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_md_file():
    """
    Test to make sure we get help if '-l' is supplied with a file path that is a simple md file.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "simple/simple.md"]

    expected_return_code = 0
    expected_output = """simple/simple.md
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_mixed_files():
    """
    Test to make sure we get help if '-l' is supplied with a file path that is a simple md file and one that isn't.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "-l",
        "only-text/simple_text_file.txt",
        "simple/simple.md",
    ]

    expected_return_code = 1
    expected_output = """"""
    expected_error = """Provided file path 'only-text/simple_text_file.txt' is not a valid file. Skipping.
"""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_globbed_files():
    """
    Test to make sure we get help if '-l' is supplied with a globbed file path that works.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "rules/md001/*.md"]

    expected_return_code = 0
    expected_output = """rules/md001/empty.md
rules/md001/front_matter_with_alternate_title.md
rules/md001/front_matter_with_no_title.md
rules/md001/front_matter_with_title.md
rules/md001/improper_atx_heading_incrementing.md
rules/md001/improper_setext_heading_incrementing.md
rules/md001/proper_atx_heading_incrementing.md
rules/md001/proper_setext_heading_incrementing.md"""
    expected_error = """"""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_non_matching_globbed_files():
    """
    Test to make sure we get help if '-l' is supplied with a globbed file path that works.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "rules/md001/z*.md"]

    expected_return_code = 1
    expected_output = """"""
    expected_error = (
        """Provided glob path 'rules/md001/z*.md' did not match any files."""
    )

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_on_directory():
    """
    Test to make sure we get help if '-l' is supplied with a directory.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "../../docs"]

    expected_return_code = 0
    expected_output = """../../docs/advanced_configuration.md
../../docs/advanced_plugins.md
../../docs/advanced_scanning.md
../../docs/developer.md
../../docs/faq.md
../../docs/rules.md"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_l_and_dash_r_on_directory():
    """
    Test to make sure we get help if '-l' and '-r' is supplied with a directory.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l", "-r", "../../docs"]

    expected_return_code = 0
    expected_output = """../../docs/advanced_configuration.md
../../docs/advanced_plugins.md
../../docs/advanced_scanning.md
../../docs/developer.md
../../docs/extensions/front-matter.md
../../docs/faq.md
../../docs/rules.md
../../docs/rules/rule_md001.md
../../docs/rules/rule_md002.md
../../docs/rules/rule_md003.md
../../docs/rules/rule_md018.md
../../docs/rules/rule_md019.md
../../docs/rules/rule_md020.md
../../docs/rules/rule_md021.md
../../docs/rules/rule_md022.md
../../docs/rules/rule_md023.md
../../docs/rules/rule_md024.md
../../docs/rules/rule_md026.md
../../docs/rules/rule_md036.md
../../docs/rules/rule_md047.md"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, cwd=scanner.resource_directory
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_paragraph_series_e_cs():
    """
    Test case:  Paragraph with code span with newline inside
    was:        test_paragraph_extra_43
    """

    # Arrange
    source_markdown = """a`code
span`a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[icode-span(1,2):code\a\n\a \aspan:`::]",
        "[text(2,6):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<code>code span</code>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_e_rh():
    """
    Test case:  Paragraph with raw HTML with newline inside
    was:        test_paragraph_extra_44
    """

    # Arrange
    source_markdown = """a<raw
html='cool'>a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[raw-html(1,2):raw\nhtml='cool']",
        "[text(2,13):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<raw\nhtml='cool'>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_e_ua():
    """
    Test case:  Paragraph with URI autolink with newline inside, renders invalid
    was:        test_paragraph_extra_45
    """

    # Arrange
    source_markdown = """a<http://www.\ngoogle.com>a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a\a<\a&lt;\ahttp://www.\ngoogle.com\a>\a&gt;\aa::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a&lt;http://www.\ngoogle.com&gt;a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_e_ea():
    """
    Test case:  Paragraph with email autolink with newline inside, renders invalid
    was:        test_paragraph_extra_46
    """

    # Arrange
    source_markdown = """a<foo@bar\n.com>a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a\a<\a&lt;\afoo@bar\n.com\a>\a&gt;\aa::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a&lt;foo@bar\n.com&gt;a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_e_em():
    """
    Test case:  Paragraph with emphasis with newline inside
    was:        test_paragraph_extra_46b
    """

    # Arrange
    source_markdown = """a*foo\nbar*a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[emphasis(1,2):1:*]",
        "[text(1,3):foo\nbar::\n]",
        "[end-emphasis(2,4)::]",
        "[text(2,5):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<em>foo\nbar</em>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide tests for source providers.
"""
import os

from pymarkdown.source_providers import FileSourceProvider, InMemorySourceProvider


def __verify_line(expected_line, actual_line):
    if expected_line is None:
        assert actual_line is None, (
            "Actual line should be None, not: '" + actual_line + "'"
        )
    else:
        assert actual_line == expected_line, (
            "Actual line:   '"
            + actual_line
            + "'\n"
            + "Expected line: '"
            + expected_line
            + "'\n"
        )


def test_source_provider_in_memory_empty_string():
    """
    Test the in memory source provider with an empty string.
    """

    # Arrange
    source_provider = InMemorySourceProvider("")
    expected_first_line = ""
    expected_second_line = None

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)


def test_source_provider_in_memory_one_line():
    """
    Test the in memory source provider with one line of input.
    """

    # Arrange
    expected_first_line = "this is the first line"
    expected_second_line = None
    source_provider = InMemorySourceProvider(expected_first_line)

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)


def test_source_provider_in_memory_two_lines():
    """
    Test the in memory source provider with two lines of input.
    """

    # Arrange
    expected_first_line = "this is the first line"
    expected_second_line = "this is the second line"
    expected_third_line = None
    source_provider = InMemorySourceProvider(
        expected_first_line + "\n" + expected_second_line
    )

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()
    actual_third_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)
    __verify_line(expected_third_line, actual_third_line)


def test_source_provider_file_empty():
    """
    Test the file source provider with an empty file.
    """

    # Arrange
    resource_directory = os.path.join(os.getcwd(), "test", "resources")
    input_file = os.path.join(resource_directory, "empty-file.txt")
    source_provider = FileSourceProvider(input_file)
    expected_first_line = ""
    expected_second_line = None

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)


def test_source_provider_file_single_line():
    """
    Test the file source provider with a single line
    """

    # Arrange
    resource_directory = os.path.join(os.getcwd(), "test", "resources")
    input_file = os.path.join(resource_directory, "single-line.txt")
    source_provider = FileSourceProvider(input_file)
    expected_first_line = "this is the first line"
    expected_second_line = None

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)


def test_source_provider_file_two_lines():
    """
    Test the file source provider with two lines of input.
    """

    # Arrange
    resource_directory = os.path.join(os.getcwd(), "test", "resources")
    input_file = os.path.join(resource_directory, "double-line.txt")
    source_provider = FileSourceProvider(input_file)
    expected_first_line = "this is the first line"
    expected_second_line = "this is the second line"
    expected_third_line = None

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()
    actual_third_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)
    __verify_line(expected_third_line, actual_third_line)


def test_source_provider_file_two_lines_with_blank_between():
    """
    Test the file source provider with two lines of input with a blank between them.
    """

    # Arrange
    resource_directory = os.path.join(os.getcwd(), "test", "resources")
    input_file = os.path.join(resource_directory, "double-line-with-blank.txt")
    source_provider = FileSourceProvider(input_file)
    expected_first_line = "this is the first line"
    expected_second_line = ""
    expected_third_line = "this is the second line"
    expected_fourth_line = None

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()
    actual_third_line = source_provider.get_next_line()
    actual_fourth_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)
    __verify_line(expected_third_line, actual_third_line)
    __verify_line(expected_fourth_line, actual_fourth_line)


def test_source_provider_file_two_lines_with_blank_between_and_trailing():
    """
    Test the file source provider with two lines of input with a blank between them and a trailing line.
    """

    # Arrange
    resource_directory = os.path.join(os.getcwd(), "test", "resources")
    input_file = os.path.join(
        resource_directory, "double-line-with-blank-and-trailing.txt"
    )
    source_provider = FileSourceProvider(input_file)
    expected_first_line = "this is the first line"
    expected_second_line = ""
    expected_third_line = "this is the second line"
    expected_fourth_line = ""
    expected_fifth_line = None

    # Act
    actual_first_line = source_provider.get_next_line()
    actual_second_line = source_provider.get_next_line()
    actual_third_line = source_provider.get_next_line()
    actual_fourth_line = source_provider.get_next_line()
    actual_fifth_line = source_provider.get_next_line()

    # Assert
    __verify_line(expected_first_line, actual_first_line)
    __verify_line(expected_second_line, actual_second_line)
    __verify_line(expected_third_line, actual_third_line)
    __verify_line(expected_fourth_line, actual_fourth_line)
    __verify_line(expected_fifth_line, actual_fifth_line)



"""
Inline processing
"""
import logging

from pymarkdown.emphasis_helper import EmphasisHelper
from pymarkdown.inline_helper import InlineHelper, InlineRequest, InlineResponse
from pymarkdown.inline_markdown_token import SpecialTextMarkdownToken, TextMarkdownToken
from pymarkdown.link_helper import LinkHelper
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines


# pylint: disable=too-few-public-methods
class InlineProcessor:
    """
    Handle the inline processing of the token stream.
    """

    __valid_inline_text_block_sequence_starts = ""
    __valid_inline_simple_text_block_sequence_starts = ""
    __inline_processing_needed = f"{EmphasisHelper.inline_emphasis}{LinkHelper.link_label_start}{LinkHelper.link_label_end}"
    __inline_character_handlers = {}
    __inline_simple_character_handlers = {}

    """
    Class to provide helper functions for parsing html.
    """

    @staticmethod
    def initialize():
        """
        Initialize the inline processor subsystem.
        """
        InlineProcessor.__inline_character_handlers = {}
        InlineProcessor.__inline_simple_character_handlers = {}
        InlineProcessor.__valid_inline_text_block_sequence_starts = (
            ParserHelper.newline_character
        )
        InlineProcessor.__valid_inline_simple_text_block_sequence_starts = (
            ParserHelper.newline_character
        )

        InlineProcessor.register_handlers(
            InlineHelper.code_span_bounds, InlineHelper.handle_inline_backtick
        )
        InlineProcessor.register_handlers(
            InlineHelper.backslash_character,
            InlineHelper.handle_inline_backslash,
            is_simple_handler=True,
        )
        InlineProcessor.register_handlers(
            InlineHelper.character_reference_start_character,
            InlineHelper.handle_character_reference,
            is_simple_handler=True,
        )
        InlineProcessor.register_handlers(
            InlineHelper.angle_bracket_start, InlineHelper.handle_angle_brackets
        )
        for i in InlineProcessor.__inline_processing_needed:
            InlineProcessor.register_handlers(
                i, InlineProcessor.__handle_inline_special_single_character
            )
        InlineProcessor.register_handlers(
            LinkHelper.image_start_sequence[0],
            InlineProcessor.__handle_inline_image_link_start_character,
        )
        for i in ParserHelper.valid_characters_to_escape():
            InlineProcessor.register_handlers(
                i, InlineProcessor.__handle_inline_control_character
            )

    @staticmethod
    def __handle_inline_control_character(inline_request):
        inline_response = InlineResponse()
        (
            inline_response.new_index,
            inline_response.new_string,
            inline_response.delta_column_number,
        ) = (
            inline_request.next_index + 1,
            f"{ParserHelper.escape_character}{inline_request.source_text[inline_request.next_index]}",
            1,
        )
        return inline_response

    @staticmethod
    def register_handlers(
        inline_character, start_token_handler, is_simple_handler=False
    ):
        """
        Register the handlers necessary to deal with token's start and end.
        """
        InlineProcessor.__inline_character_handlers[
            inline_character
        ] = start_token_handler
        InlineProcessor.__valid_inline_text_block_sequence_starts += inline_character
        if is_simple_handler:
            InlineProcessor.__inline_simple_character_handlers[
                inline_character
            ] = start_token_handler
            InlineProcessor.__valid_inline_simple_text_block_sequence_starts += (
                inline_character
            )

    @staticmethod
    # pylint: disable=too-many-branches, too-many-statements, too-many-nested-blocks
    def parse_inline(coalesced_results):
        """
        Parse and resolve any inline elements.
        """
        POGGER.info("coalesced_results")
        POGGER.info("-----")
        for next_token in coalesced_results:
            POGGER.info(">>$<<", next_token)
        POGGER.info("-----")

        coalesced_stack, coalesced_list, current_token = [], [], coalesced_results[0]
        coalesced_list.extend(coalesced_results[0:1])

        POGGER.debug("STACK?:$", current_token)
        if current_token.is_container:
            POGGER.debug("STACK:$", coalesced_stack)
            coalesced_stack.append(current_token)
            POGGER.debug("STACK-ADD:$", current_token)
            POGGER.debug("STACK:$", coalesced_stack)

        for coalesce_index in range(1, len(coalesced_results)):
            if coalesced_results[coalesce_index].is_text and (
                coalesced_list[-1].is_paragraph
                or coalesced_list[-1].is_setext_heading
                or coalesced_list[-1].is_atx_heading
                or coalesced_list[-1].is_code_block
            ):
                POGGER.info("coalesced_results:$<", coalesced_list[-1])
                if coalesced_list[-1].is_code_block:
                    encoded_text = InlineHelper.append_text(
                        "", coalesced_results[coalesce_index].token_text
                    )
                    line_number_delta, new_column_number = (
                        0,
                        coalesced_list[-1].column_number,
                    )
                    if coalesced_list[-1].is_fenced_code_block:
                        line_number_delta, new_column_number = 1, 1

                        POGGER.info("coalesced_stack:$<", coalesced_stack)
                        if coalesced_stack:
                            assert coalesced_stack[-1].leading_spaces
                            split_leading_spaces = coalesced_stack[
                                -1
                            ].leading_spaces.split(ParserHelper.newline_character)
                            new_column_number += (
                                (len(split_leading_spaces[1]))
                                if len(split_leading_spaces) >= 2
                                else (len(split_leading_spaces[0]))
                            )
                        else:
                            leading_whitespace = coalesced_results[
                                coalesce_index
                            ].extracted_whitespace
                            POGGER.debug(">>$<<", coalesced_results[coalesce_index])
                            assert (
                                ParserHelper.newline_character not in leading_whitespace
                            )
                            POGGER.info(
                                "leading_whitespace:$<",
                                leading_whitespace,
                            )
                            leading_whitespace = ParserHelper.remove_all_from_text(
                                leading_whitespace
                            )
                            POGGER.info("leading_whitespace:$<", leading_whitespace)
                            new_column_number += len(leading_whitespace)
                    processed_tokens = [
                        TextMarkdownToken(
                            encoded_text,
                            coalesced_results[coalesce_index].extracted_whitespace,
                            line_number=coalesced_list[-1].line_number
                            + line_number_delta,
                            column_number=new_column_number,
                        )
                    ]
                    POGGER.debug(
                        "new Text>>$>>",
                        processed_tokens,
                    )
                elif coalesced_list[-1].is_setext_heading:
                    combined_text, combined_whitespace_text = (
                        coalesced_results[coalesce_index].token_text,
                        coalesced_results[coalesce_index].extracted_whitespace.replace(
                            ParserHelper.tab_character, "    "
                        ),
                    )
                    POGGER.debug(
                        "combined_text>>$",
                        combined_text,
                    )
                    POGGER.debug(
                        "combined_whitespace_text>>$",
                        combined_whitespace_text,
                    )
                    processed_tokens = InlineProcessor.__process_inline_text_block(
                        coalesced_results[coalesce_index].token_text.replace(
                            ParserHelper.tab_character, "    "
                        ),
                        whitespace_to_recombine=combined_whitespace_text,
                        is_setext=True,
                        para_space=coalesced_results[
                            coalesce_index
                        ].extracted_whitespace,
                        line_number=coalesced_results[coalesce_index].line_number,
                        column_number=coalesced_results[coalesce_index].column_number,
                    )
                    POGGER.debug(
                        "processed_tokens>>$",
                        processed_tokens,
                    )
                elif coalesced_list[-1].is_atx_heading:
                    POGGER.debug("atx-block>>$<<", coalesced_results[coalesce_index])
                    POGGER.debug(
                        "atx-block-text>>$<<",
                        coalesced_results[coalesce_index].token_text,
                    )
                    POGGER.debug(
                        "atx-block-ws>>$<<",
                        coalesced_results[coalesce_index].extracted_whitespace,
                    )
                    processed_tokens = InlineProcessor.__process_inline_text_block(
                        coalesced_results[coalesce_index].token_text,
                        coalesced_results[coalesce_index].extracted_whitespace,
                        line_number=coalesced_results[coalesce_index].line_number,
                        column_number=coalesced_results[coalesce_index].column_number
                        + len(coalesced_results[coalesce_index].extracted_whitespace)
                        + coalesced_list[-1].hash_count,
                    )
                else:
                    assert coalesced_list[-1].is_paragraph
                    POGGER.debug(
                        ">>before_add_ws>>$>>add>>$>>",
                        coalesced_list[-1],
                        coalesced_results[coalesce_index].extracted_whitespace,
                    )
                    coalesced_list[-1].add_whitespace(
                        coalesced_results[coalesce_index].extracted_whitespace.replace(
                            ParserHelper.tab_character, "    "
                        )
                    )
                    POGGER.debug(">>after_add_ws>>$", coalesced_list[-1])
                    processed_tokens = InlineProcessor.__process_inline_text_block(
                        coalesced_results[coalesce_index].token_text.replace(
                            ParserHelper.tab_character, "    "
                        ),
                        is_para=True,
                        para_space=coalesced_results[
                            coalesce_index
                        ].extracted_whitespace,
                        line_number=coalesced_results[coalesce_index].line_number,
                        column_number=coalesced_results[coalesce_index].column_number,
                        para_owner=coalesced_list[-1],
                    )
                coalesced_list.extend(processed_tokens)
            else:
                coalesced_list.append(coalesced_results[coalesce_index])

            current_token = coalesced_results[coalesce_index]
            POGGER.debug("STACK?:$", current_token)
            if current_token.is_container and not current_token.is_new_list_item:
                POGGER.debug("STACK:$", coalesced_stack)
                coalesced_stack.append(current_token)
                POGGER.debug("STACK-ADD:$", current_token)
                POGGER.debug("STACK:$", coalesced_stack)

            elif current_token.is_list_end or current_token.is_block_quote_end:
                POGGER.debug("STACK:$", coalesced_stack)
                del coalesced_stack[-1]
                POGGER.debug(
                    "STACK-REMOVE:$",
                    current_token,
                )
                POGGER.debug("STACK:$", coalesced_stack)
        return coalesced_list

    # pylint: enable=too-many-branches, too-many-statements, too-many-nested-blocks

    @staticmethod
    def __handle_inline_special_single_character(inline_request):
        return InlineProcessor.__handle_inline_special(
            inline_request.source_text,
            inline_request.next_index,
            inline_request.inline_blocks,
            1,
            inline_request.remaining_line,
            inline_request.current_string_unresolved,
            inline_request.line_number,
            inline_request.column_number,
            inline_request.para_owner,
        )

    @staticmethod
    def __handle_inline_image_link_start_character(inline_request):
        if ParserHelper.are_characters_at_index(
            inline_request.source_text,
            inline_request.next_index,
            LinkHelper.image_start_sequence,
        ):
            inline_response = InlineProcessor.__handle_inline_special(
                inline_request.source_text,
                inline_request.next_index,
                inline_request.inline_blocks,
                2,
                inline_request.remaining_line,
                inline_request.current_string_unresolved,
                inline_request.line_number,
                inline_request.column_number,
                inline_request.para_owner,
            )
            assert not inline_response.consume_rest_of_line
        else:
            inline_response = InlineResponse()
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.delta_column_number,
            ) = (LinkHelper.image_start_sequence[0], inline_request.next_index + 1, 1)
        return inline_response

    # pylint: disable=too-many-arguments, too-many-locals, too-many-statements, too-many-branches, too-many-nested-blocks
    @staticmethod
    def __handle_inline_special(
        source_text,
        next_index,
        inline_blocks,
        special_length,
        remaining_line,
        current_string_unresolved,
        line_number,
        column_number,
        para_owner,
    ):
        """
        Handle the collection of special inline characters for later processing.
        """
        (
            preceding_two,
            following_two,
            new_token,
            repeat_count,
            is_active,
            delta_line,
            consume_rest_of_line,
            remaining_line_size,
        ) = (None, None, None, 1, True, 0, False, len(remaining_line))
        POGGER.debug(">>column_number>>$<<", column_number)
        POGGER.debug(">>remaining_line>>$<<", remaining_line)
        column_number += remaining_line_size
        POGGER.debug(">>column_number>>$<<", column_number)
        special_sequence = source_text[next_index : next_index + special_length]
        if special_length == 1 and special_sequence in EmphasisHelper.inline_emphasis:
            repeat_count, new_index = ParserHelper.collect_while_character(
                source_text, next_index, special_sequence
            )
            special_sequence, preceding_two, following_two = (
                source_text[next_index:new_index],
                source_text[max(0, next_index - 2) : next_index],
                source_text[new_index : min(len(source_text), new_index + 2)],
            )
        else:
            if special_sequence[0] == LinkHelper.link_label_end:
                POGGER.debug(
                    "POSSIBLE LINK CLOSE_FOUND($)>>$>>",
                    special_length,
                    special_sequence,
                )
                POGGER.debug(
                    ">>inline_blocks>>$<<",
                    inline_blocks,
                )
                POGGER.debug(
                    ">>remaining_line>>$<<",
                    remaining_line,
                )
                POGGER.debug(
                    ">>current_string_unresolved>>$<<",
                    current_string_unresolved,
                )
                POGGER.debug(
                    ">>source_text>>$<<",
                    source_text[next_index:],
                )
                POGGER.debug("")
                old_inline_blocks_count, old_inline_blocks_last_token = (
                    len(inline_blocks),
                    inline_blocks[-1] if inline_blocks else None,
                )
                (
                    new_index,
                    is_active,
                    new_token,
                    consume_rest_of_line,
                ) = LinkHelper.look_for_link_or_image(
                    inline_blocks,
                    source_text,
                    next_index,
                    remaining_line,
                    current_string_unresolved,
                    InlineProcessor.__process_simple_inline_fn,
                )
                POGGER.debug(">>next_index>>$<<", next_index)
                POGGER.debug(">>new_index>>$<<", new_index)
                POGGER.debug(">>inline_blocks>>$<<", inline_blocks)
                POGGER.debug(">>new_token>>$<<", new_token)
                POGGER.debug(">>source_text>>$<<", source_text[new_index:])
                POGGER.debug(">>consume_rest_of_line>>$<<", consume_rest_of_line)
                POGGER.debug(">>old_inline_blocks_count>>$<<", old_inline_blocks_count)

                new_inline_blocks_count = len(inline_blocks)
                POGGER.debug(">>new_inline_blocks_count>>$<<", new_inline_blocks_count)

                if (
                    new_token
                    or old_inline_blocks_count != new_inline_blocks_count
                    or (
                        inline_blocks
                        and old_inline_blocks_last_token != inline_blocks[-1]
                    )
                ):
                    if inline_blocks[-1].is_inline_image:
                        repeat_count = (new_index - next_index) + remaining_line_size
                        (
                            delta_line,
                            repeat_count,
                        ) = InlineProcessor.__calculate_link_and_image_deltas(
                            para_owner, inline_blocks[-1], delta_line, repeat_count
                        )
                        POGGER.debug(">>delta_line>>$<<", delta_line)
                        POGGER.debug(">>repeat_count>>$<<", repeat_count)
                    elif new_token and new_token.is_inline_link_end:
                        POGGER.debug(
                            ">>new_token.start_markdown_token>>$<<",
                            new_token.start_markdown_token,
                        )
                        assert new_token.start_markdown_token
                        repeat_count = new_index - next_index
                        POGGER.debug(">>delta_line>>$<<", delta_line)
                        POGGER.debug(">>repeat_count>>$<<", repeat_count)
                        (
                            delta_line,
                            repeat_count,
                        ) = InlineProcessor.__calculate_link_and_image_deltas(
                            para_owner,
                            new_token.start_markdown_token,
                            delta_line,
                            repeat_count,
                        )
                        POGGER.debug(">>delta_line>>$<<", delta_line)
                        POGGER.debug(">>repeat_count>>$<<", repeat_count)
                    else:
                        repeat_count = new_index - next_index
                        POGGER.debug(">>repeat_count>>$<<", repeat_count)
            else:
                repeat_count, new_index = special_length, next_index + special_length

        if not new_token:
            POGGER.debug(">>create>>$,$<<", line_number, column_number)
            new_token = SpecialTextMarkdownToken(
                special_sequence,
                repeat_count,
                preceding_two,
                following_two,
                is_active,
                line_number,
                column_number,
            )

        POGGER.debug(">>delta_line>>$<<", delta_line)
        POGGER.debug(">>repeat_count>>$<<", repeat_count)
        inline_response = InlineResponse()
        (
            inline_response.new_string,
            inline_response.new_index,
            inline_response.new_tokens,
            inline_response.consume_rest_of_line,
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ("", new_index, [new_token], consume_rest_of_line, delta_line, repeat_count)
        return inline_response

    # pylint: enable=too-many-arguments, too-many-locals, too-many-statements, too-many-branches, too-many-nested-blocks

    @staticmethod
    def __process_simple_inline_fn(source_text):
        """
        Handle a simple processing of inline text for simple replacements.
        """

        POGGER.debug(">>source_text>>$", source_text)
        start_index = 0
        POGGER.debug(
            ">>__valid_inline_simple_text_block_sequence_starts>>$",
            InlineProcessor.__valid_inline_simple_text_block_sequence_starts,
        )
        next_index = ParserHelper.index_any_of(
            source_text,
            InlineProcessor.__valid_inline_simple_text_block_sequence_starts,
            start_index,
        )
        processed_parts = []
        POGGER.debug(">>next_index>>$", next_index)
        while next_index != -1:
            processed_parts.append(source_text[start_index:next_index])
            inline_request = InlineRequest(source_text, next_index)
            if source_text[next_index] in InlineProcessor.__inline_character_handlers:
                POGGER.debug(
                    "handler(before)>>$<<",
                    source_text[next_index],
                )
                POGGER.debug("current_string_unresolved>>$<<", processed_parts)
                proc_fn = InlineProcessor.__inline_character_handlers[
                    source_text[next_index]
                ]
                inline_response = proc_fn(inline_request)
                processed_parts.append(inline_response.new_string)
                POGGER.debug("handler(after)>>$<<", source_text[next_index])
                POGGER.debug(
                    "delta_line_number>>$<<", inline_response.delta_line_number
                )
                POGGER.debug("delta_column>>$<<", inline_response.delta_column_number)
                start_index = inline_response.new_index
            else:
                processed_parts.append(ParserHelper.newline_character)
                start_index = next_index + 1
            next_index = ParserHelper.index_any_of(
                source_text,
                InlineProcessor.__valid_inline_simple_text_block_sequence_starts,
                start_index,
            )

        processed_parts.append(source_text[start_index:])
        return "".join(processed_parts)

    @staticmethod
    def __calculate_full_deltas(current_token, para_owner, delta_line, repeat_count):
        newline_count = ParserHelper.count_newlines_in_text(current_token.ex_label)
        if newline_count:
            POGGER.debug(">>ex_label")
            delta_line += newline_count
            if para_owner:
                para_owner.rehydrate_index += newline_count
            POGGER.debug("full>>ex_label>>newline_count>>$", newline_count)

            last_line_of_label = ParserHelper.calculate_last_line(
                current_token.ex_label
            )
            repeat_count = -(len(last_line_of_label) + 2)
        return delta_line, repeat_count

    # pylint: disable=too-many-arguments, too-many-branches, too-many-statements
    @staticmethod
    def __calculate_inline_deltas(
        current_token,
        para_owner,
        split_paragraph_lines,
        delta_line,
        repeat_count,
    ):
        last_spaces, active_link_uri, active_link_title = (
            "",
            current_token.active_link_uri,
            current_token.active_link_title,
        )

        link_part_lengths = [0] * 5
        link_part_lengths[0] = len(active_link_uri) + len(
            current_token.before_title_whitespace
        )
        if current_token.inline_title_bounding_character:
            link_part_lengths[1] = 1
            link_part_lengths[2] = len(active_link_title) + 1
            link_part_lengths[3] = len(current_token.after_title_whitespace)

        POGGER.debug(">>link_part_lengths>>$<<", link_part_lengths)

        link_part_index, total_newlines = -2, 0
        newline_count = ParserHelper.count_newlines_in_text(
            current_token.text_from_blocks
        )
        if newline_count:
            link_part_index, total_newlines = -1, total_newlines + newline_count

        newline_count = ParserHelper.count_newlines_in_text(
            current_token.before_link_whitespace
        )
        if newline_count:
            POGGER.debug(">>before_link_whitespace")
            link_part_index, delta_line, last_spaces = (
                0,
                delta_line + newline_count,
                current_token.before_link_whitespace[:],
            )

        newline_count = ParserHelper.count_newlines_in_text(
            current_token.before_title_whitespace
        )
        if newline_count:
            POGGER.debug(">>before_title_whitespace")
            link_part_index, delta_line, last_spaces = (
                1,
                delta_line + newline_count,
                current_token.before_title_whitespace[:],
            )

        newline_count = ParserHelper.count_newlines_in_text(active_link_title)
        if newline_count:
            POGGER.debug(">>active_link_title")
            _, delta_column_number = ParserHelper.calculate_deltas(active_link_title)
            link_part_index, delta_line, last_spaces, link_part_lengths[2] = (
                2,
                delta_line + newline_count,
                "",
                -delta_column_number,
            )

        newline_count = ParserHelper.count_newlines_in_text(
            current_token.after_title_whitespace
        )
        if newline_count:
            POGGER.debug(">>after_title_whitespace")
            link_part_index, delta_line, last_spaces = (
                4,
                delta_line + newline_count,
                current_token.after_title_whitespace[:],
            )

        POGGER.debug(">>link_part_index>>$<<", link_part_index)
        POGGER.debug(">>delta_line>>$<<", delta_line)

        if para_owner:
            para_owner.rehydrate_index += total_newlines + delta_line
        if link_part_index >= 0:
            link_part_lengths[4] = (
                len(split_paragraph_lines[para_owner.rehydrate_index])
                if split_paragraph_lines
                else len(ParserHelper.calculate_last_line(last_spaces))
            )
            link_part_lengths[:link_part_index] = [0] * link_part_index
            repeat_count = -(2 + sum(link_part_lengths))
            POGGER.debug(">>link_part_lengths>>$<<", link_part_lengths)
            POGGER.debug(">>repeat_count>>$<<", delta_line)
        return delta_line, repeat_count

    # pylint: enable=too-many-arguments, too-many-branches, too-many-statements

    @staticmethod
    def __calculate_shortcut_collapsed_deltas(current_token, delta_line, repeat_count):
        """
        Tests test_reference_links_extra_03jx and test_reference_links_extra_03ja added
        to ensure that this is correct.  Those tests confirm that any newlines in the
        label are already accounted for, and as such, do not require any further
        modifications.
        """
        _ = current_token
        return delta_line, repeat_count

    @staticmethod
    def __calculate_link_and_image_deltas(
        para_owner, current_token, delta_line, repeat_count
    ):
        POGGER.debug(">>delta_line>>$<<", delta_line)
        POGGER.debug(">>repeat_count>>$<<", repeat_count)

        if ParserHelper.newline_character in str(current_token):
            POGGER.debug(">>para_owner>>$<<", para_owner)
            split_paragraph_lines = None
            if para_owner:
                POGGER.debug(
                    ">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index
                )
                split_paragraph_lines = para_owner.extracted_whitespace.split(
                    ParserHelper.newline_character
                )
                POGGER.debug(">>split_paragraph_lines>>$<<", split_paragraph_lines)

            POGGER.debug(">>current_token.label_type>>$<<", current_token.label_type)
            if current_token.label_type == "inline":
                delta_line, repeat_count = InlineProcessor.__calculate_inline_deltas(
                    current_token,
                    para_owner,
                    split_paragraph_lines,
                    delta_line,
                    repeat_count,
                )
            elif current_token.label_type == "full":
                delta_line, repeat_count = InlineProcessor.__calculate_full_deltas(
                    current_token, para_owner, delta_line, repeat_count
                )
            else:
                assert current_token.label_type in (
                    "shortcut",
                    "collapsed",
                ), f"Label type '{current_token.label_type}' not handled."
                (
                    delta_line,
                    repeat_count,
                ) = InlineProcessor.__calculate_shortcut_collapsed_deltas(
                    current_token, delta_line, repeat_count
                )

        POGGER.debug(">>delta_line>>$<<repeat_count>>$<<", delta_line, repeat_count)
        return delta_line, repeat_count

    @staticmethod  # noqa: C901
    # pylint: disable=too-many-statements, too-many-locals, too-many-arguments, too-many-branches
    def __process_inline_text_block(
        source_text,
        starting_whitespace="",
        whitespace_to_recombine=None,
        is_setext=False,
        is_para=False,
        para_space=None,
        line_number=0,
        column_number=0,
        para_owner=None,
    ):
        """
        Process a text block for any inline items.

        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """

        inline_blocks, start_index = [], 0
        # POGGER.debug(
        #    "__process_inline_text_block>>source_text>>$>",
        #    source_text,
        # )
        # POGGER.debug(
        #    "__process_inline_text_block>>starting_whitespace>>$>",
        #    starting_whitespace,
        # )
        # POGGER.debug(
        #     "__process_inline_text_block>>whitespace_to_recombine>>$>",
        #     whitespace_to_recombine,
        # )
        # POGGER.debug(
        #     "__process_inline_text_block>>line_number>>$>",
        #     line_number,
        # )
        # POGGER.debug(
        #     "__process_inline_text_block>>column_number>>$>",
        #     column_number,
        # )
        if whitespace_to_recombine:
            source_text, _ = ParserHelper.recombine_string_with_whitespace(
                source_text, whitespace_to_recombine
            )
        # POGGER.debug(
        #     "__process_inline_text_block>>source_text>>$",
        #     source_text,
        # )

        (
            last_line_number,
            last_column_number,
            current_string,
            current_string_unresolved,
            end_string,
            inline_response,
            fold_space,
        ) = (line_number, column_number, "", "", "", InlineResponse(), None)

        # POGGER.debug(
        #     ">>Token_start>>$,$<<",
        #     last_line_number,
        #     last_column_number,
        # )
        # POGGER.debug("__process_inline_text_block>>is_para>>$", is_para)
        if is_para or is_setext:
            fold_space = para_space.split(ParserHelper.newline_character)
        # POGGER.debug("__process_inline_text_block>>fold_space>>$", fold_space)

        # POGGER.debug(
        #     "starts>$<", InlineProcessor.__valid_inline_text_block_sequence_starts
        # )
        # POGGER.debug("look>$<", source_text[start_index:])
        next_index = ParserHelper.index_any_of(
            source_text,
            InlineProcessor.__valid_inline_text_block_sequence_starts,
            start_index,
        )
        # POGGER.debug("__process_inline_text_block>>is_setext>>$", is_setext)
        # POGGER.debug(
        #     "__process_inline_text_block>>$>>$",
        #     source_text,
        #     start_index,
        # )
        while next_index != -1:

            # POGGER.debug(
            #     "\n\n>>Token_start>>$,$<<",
            #     last_line_number,
            #     last_column_number,
            # )
            # POGGER.debug(">>inline_blocks>>$<<", inline_blocks)
            # POGGER.debug(">>current_string>>$<<", current_string)
            # POGGER.debug(">>current_string_unresolved>>$<<", current_string_unresolved)
            # POGGER.debug(">>current_string_unresolved>>$<<", current_string_unresolved)
            # POGGER.debug(">>end_string>>$<<", end_string)
            # POGGER.debug(
            #     ">>source_text[]>>$<<$<<",
            #     source_text[next_index],
            #     source_text[next_index:],
            # )

            inline_response.clear_fields()
            (
                reset_current_string,
                whitespace_to_add,
                was_new_line,
                was_column_number_reset,
                remaining_line,
                old_inline_blocks_count,
                old_inline_blocks_last_token,
            ) = (
                False,
                None,
                False,
                False,
                source_text[start_index:next_index],
                len(inline_blocks),
                inline_blocks[-1] if inline_blocks else None,
            )

            # POGGER.debug("__process_inline_text_block>>$>>$", start_index, next_index)
            # POGGER.debug(
            #     "__process_inline_text_block>>$<<", source_text[start_index:next_index]
            # )
            inline_request = InlineRequest(
                source_text,
                next_index,
                inline_blocks,
                remaining_line,
                current_string_unresolved,
                line_number,
                column_number,
                para_owner,
            )
            if source_text[next_index] in InlineProcessor.__inline_character_handlers:
                # POGGER.debug(
                #     "handler(before)>>$<<",
                #     source_text[next_index],
                # )
                # POGGER.debug(
                #     "current_string_unresolved>>$<<",
                #     current_string_unresolved,
                # )
                # POGGER.debug("remaining_line>>$<<", remaining_line)
                # POGGER.debug("line_number>>$<<", line_number)
                # POGGER.debug("column_number>>$<<", column_number)
                proc_fn = InlineProcessor.__inline_character_handlers[
                    source_text[next_index]
                ]
                inline_response = proc_fn(inline_request)
                # POGGER.debug(
                #     "handler(after)>>$<<",
                #     source_text[next_index],
                # )
                # POGGER.debug(
                #     "delta_line_number>>$<<", inline_response.delta_line_number
                # )
                # POGGER.debug("delta_column>>$<<", inline_response.delta_column_number)

                line_number += inline_response.delta_line_number
                if inline_response.delta_column_number < 0:
                    column_number, was_column_number_reset = (
                        -inline_response.delta_column_number,
                        True,
                    )
                else:
                    column_number += inline_response.delta_column_number
                # POGGER.debug("handler(after)>>$,$<<", line_number, column_number)
                # POGGER.debug(
                #     "handler(after)>>new_tokens>>$<<",
                #     inline_response.new_tokens,
                # )
            else:
                assert source_text[next_index] == ParserHelper.newline_character
                # POGGER.debug(
                #     "end_string(before)>>$<<",
                #     end_string,
                # )
                (
                    inline_response.new_string,
                    whitespace_to_add,
                    inline_response.new_index,
                    inline_response.new_tokens,
                    remaining_line,
                    end_string,
                    current_string,
                ) = InlineHelper.handle_line_end(
                    next_index,
                    remaining_line,
                    end_string,
                    current_string,
                    inline_blocks,
                    is_setext,
                    line_number,
                    column_number,
                )
                # POGGER.debug("2<<end_string<<$<<", end_string)
                # POGGER.debug(
                #     "handle_line_end>>new_tokens>>$<<",
                #     inline_response.new_tokens,
                # )
                if not inline_response.new_tokens:
                    # POGGER.debug("ws")
                    end_string = InlineProcessor.__add_recombined_whitespace(
                        bool(whitespace_to_recombine),
                        source_text,
                        inline_response,
                        end_string,
                        is_setext,
                    )
                    # POGGER.debug(
                    #     "3<<end_string<<$<<",
                    #     end_string,
                    # )
                    # POGGER.debug("ws>$<", end_string)
                # POGGER.debug(
                #     "handle_line_end>>$<<", source_text[inline_response.new_index :]
                # )
                # POGGER.debug(
                #     "end_string(after)>>$<<",
                #     end_string,
                # )
                was_new_line = True
                if para_owner:
                    para_owner.rehydrate_index += 1

            # POGGER.debug(
            #     "new_string-->$<--",
            #     inline_response.new_string,
            # )
            # POGGER.debug("new_index-->$<--", inline_response.new_index)
            # POGGER.debug(
            #     "new_tokens-->$<--",
            #     inline_response.new_tokens,
            # )
            # POGGER.debug(
            #     "new_string_unresolved-->$<--",
            #     inline_response.new_string_unresolved,
            # )
            # POGGER.debug(
            #     "consume_rest_of_line-->$<--",
            #     inline_response.consume_rest_of_line,
            # )
            # POGGER.debug(
            #     "original_string-->$<--",
            #     inline_response.original_string,
            # )

            if inline_response.consume_rest_of_line:
                # POGGER.debug("consume_rest_of_line>>$<", remaining_line)
                (
                    inline_response.new_string,
                    inline_response.new_tokens,
                    reset_current_string,
                    remaining_line,
                    end_string,
                ) = ("", None, True, "", None)
                # POGGER.debug(
                #     "9<<end_string<<$<<",
                #     end_string,
                # )
            else:
                # POGGER.debug("append_rest_of_line>>rem>>$<", remaining_line)
                # POGGER.debug("append_rest_of_line>>cur>>$<", current_string)
                # POGGER.debug(
                #     "append_rest_of_line>>cur_un>>$<", current_string_unresolved
                # )
                current_string, current_string_unresolved = (
                    InlineHelper.append_text(
                        current_string,
                        remaining_line,
                    ),
                    InlineHelper.append_text(current_string_unresolved, remaining_line),
                )

            # POGGER.debug(
            #     "current_string>>$<<",
            #     current_string,
            # )
            # POGGER.debug(
            #     "current_string_unresolved>>$<<",
            #     current_string_unresolved,
            # )
            # POGGER.debug(
            #     "inline_blocks>>$<<",
            #     inline_blocks,
            # )
            # POGGER.debug(
            #     "inline_response.new_tokens>>$<<",
            #     inline_response.new_tokens,
            # )
            # POGGER.debug(
            #     "starting_whitespace>>$<<",
            #     starting_whitespace,
            # )
            if inline_response.new_tokens:
                if current_string:
                    # POGGER.debug(">>>text1")
                    inline_blocks.append(
                        TextMarkdownToken(
                            current_string,
                            starting_whitespace,
                            end_whitespace=end_string,
                            line_number=last_line_number,
                            column_number=last_column_number,
                        )
                    )
                    # POGGER.debug("new Text>>$>>", inline_blocks)
                    reset_current_string, starting_whitespace, end_string = (
                        True,
                        "",
                        None,
                    )
                    # POGGER.debug(
                    #     "4<<end_string<<$<<",
                    #     end_string,
                    # )
                elif starting_whitespace:
                    # POGGER.debug(">>>starting whitespace")
                    inline_blocks.append(
                        TextMarkdownToken(
                            "",
                            ParserHelper.create_replace_with_nothing_marker(
                                starting_whitespace
                            ),
                            line_number=last_line_number,
                            column_number=last_column_number,
                        )
                    )
                    # POGGER.debug("new Text>>$>>", inline_blocks)
                    starting_whitespace = ""

                inline_blocks.extend(inline_response.new_tokens)

            # POGGER.debug(
            #     "l/c(before)>>$,$<<",
            #     line_number,
            #     column_number,
            # )
            if was_new_line:
                # POGGER.debug("l/c(before)>>newline")
                line_number, column_number = line_number + 1, 1
                assert fold_space
                # POGGER.debug("fold_space(before)>>$<<", fold_space)
                fold_space = fold_space[1:]
                # POGGER.debug("fold_space(after)>>$<<", fold_space)
                column_number += len(fold_space[0])

            elif not was_column_number_reset:
                column_number += len(remaining_line)
            # POGGER.debug(
            #     "l/c(after)>>$,$<<",
            #     line_number,
            #     column_number,
            # )

            # POGGER.debug(
            #     "starting_whitespace>>$<<",
            #     starting_whitespace,
            # )
            # POGGER.debug(
            #     "inline_blocks>>$<<",
            #     inline_blocks,
            # )
            # POGGER.debug("reset_current_string>>$<<", reset_current_string)

            if reset_current_string:
                current_string, current_string_unresolved = "", ""
            # POGGER.debug("pos>>$,$<<", line_number, column_number)
            # POGGER.debug("last>>$,$<<", last_line_number, last_column_number)

            inline_blocks_size = len(inline_blocks)
            # POGGER.debug(
            #     "old>>$>>now>>$<<",
            #     old_inline_blocks_count,
            #     inline_blocks_size,
            # )
            if old_inline_blocks_count != inline_blocks_size or (
                old_inline_blocks_last_token
                and old_inline_blocks_last_token != inline_blocks[-1]
            ):
                last_line_number, last_column_number = line_number, column_number
            # POGGER.debug("last>>$,$<<", last_line_number, last_column_number)
            # POGGER.debug(
            #     ">>Token_start>>$,$<<",
            #     last_line_number,
            #     last_column_number,
            # )

            # POGGER.debug("5<<end_string<<$<<", end_string)
            (
                start_index,
                next_index,
                end_string,
                current_string,
                current_string_unresolved,
            ) = InlineProcessor.__complete_inline_loop(
                source_text,
                inline_response.new_index,
                end_string,
                whitespace_to_add,
                current_string,
                current_string_unresolved,
                inline_response.new_string_unresolved,
                inline_response.new_string,
                inline_response.original_string,
            )
            # POGGER.debug("6<<end_string<<$<<", end_string)
            # POGGER.debug(
            #     "<<current_string<<$<<",
            #     current_string,
            # )
            # POGGER.debug(
            #     "<<current_string_unresolved<<$<<",
            #     current_string_unresolved,
            # )

        # POGGER.debug("<<__complete_inline_block_processing<<")
        # POGGER.debug(
        #     "<<__complete_inline_block_processing<<end_string<<$<<",
        #     end_string,
        # )
        return InlineProcessor.__complete_inline_block_processing(
            inline_blocks,
            source_text,
            start_index,
            current_string,
            end_string,
            starting_whitespace,
            is_setext,
            line_number=last_line_number,
            column_number=last_column_number,
        )

    # pylint: enable=too-many-statements, too-many-locals, too-many-arguments, too-many-branches

    @staticmethod
    def __add_recombined_whitespace(
        did_recombine, source_text, inline_response, end_string, is_setext
    ):

        POGGER.debug("__arw>>did_recombine>>$>>", did_recombine)
        POGGER.debug(
            "__arw>>end_string>>$>>",
            end_string,
        )
        if did_recombine:
            POGGER.debug(
                "__arw>>source_text>>$>>",
                source_text,
            )
            new_index, extracted_whitespace = ParserHelper.extract_whitespace(
                source_text, inline_response.new_index
            )
            POGGER.debug("__arw>>$>>", source_text[0 : inline_response.new_index])
            POGGER.debug("__arw>>$>>", source_text[inline_response.new_index :])
            POGGER.debug(
                "__arw>>extracted_whitespace>>$>>",
                extracted_whitespace,
            )
            if extracted_whitespace:
                inline_response.new_index = new_index
                assert end_string is not None
                assert is_setext
                end_string = f"{end_string}{extracted_whitespace}{ParserHelper.whitespace_split_character}"
                POGGER.debug(
                    "__arw>>end_string>>$>>",
                    end_string,
                )
        return end_string

    @staticmethod
    # pylint: disable=too-many-arguments
    def __complete_inline_loop(
        source_text,
        new_index,
        end_string,
        whitespace_to_add,
        current_string,
        current_string_unresolved,
        new_string_unresolved,
        new_string,
        original_string,
    ):
        POGGER.debug(
            "__complete_inline_loop--current_string>>$>>",
            current_string,
        )
        POGGER.debug(
            "__complete_inline_loop--new_string>>$>>",
            new_string,
        )
        POGGER.debug(
            "__complete_inline_loop--new_string_unresolved>>$>>",
            new_string_unresolved,
        )
        POGGER.debug(
            "__complete_inline_loop--original_string>>$>>",
            original_string,
        )

        POGGER.debug(
            "__complete_inline_loop--current_string>>$>>",
            current_string,
        )
        POGGER.debug(
            "__complete_inline_loop--end_string>>$>>",
            end_string,
        )
        POGGER.debug(
            "__complete_inline_loop--whitespace_to_add>>$>>",
            whitespace_to_add,
        )
        if original_string is not None:
            assert not new_string_unresolved or new_string_unresolved == original_string
            current_string = f"{current_string}{ParserHelper.create_replacement_markers(original_string, InlineHelper.append_text('', new_string))}"
        else:
            current_string = InlineHelper.append_text(current_string, new_string)
        POGGER.debug(
            "__complete_inline_loop--current_string>>$>>",
            current_string,
        )

        POGGER.debug(
            "new_string_unresolved>>$>>",
            new_string_unresolved,
        )
        if new_string_unresolved:
            current_string_unresolved = (
                f"{current_string_unresolved}{new_string_unresolved}"
            )
        else:
            current_string_unresolved = InlineHelper.append_text(
                current_string_unresolved, new_string
            )

        POGGER.debug(
            "__complete_inline_loop--current_string_unresolved>>$>>",
            current_string_unresolved,
        )

        start_index = new_index
        next_index = ParserHelper.index_any_of(
            source_text,
            InlineProcessor.__valid_inline_text_block_sequence_starts,
            start_index,
        )
        return (
            start_index,
            next_index,
            end_string,
            current_string,
            current_string_unresolved,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    # pylint: disable=too-many-arguments
    def __complete_inline_block_processing(
        inline_blocks,
        source_text,
        start_index,
        current_string,
        end_string,
        starting_whitespace,
        is_setext,
        line_number,
        column_number,
    ):
        have_processed_once = len(inline_blocks) != 0 or start_index != 0

        POGGER.debug("__cibp>inline_blocks>$<", inline_blocks)
        POGGER.debug("__cibp>source_text>$<", source_text)
        POGGER.debug("__cibp>start_index>$<", start_index)
        POGGER.debug("__cibp>current_string>$<", current_string)
        POGGER.debug("__cibp>end_string>$<", end_string)
        POGGER.debug(
            "__cibp>starting_whitespace>$<",
            starting_whitespace,
        )
        POGGER.debug("__cibp>is_setext>$<", is_setext)
        POGGER.debug("__cibp>line_number>$<", line_number)
        POGGER.debug("__cibp>column_number>$<", column_number)

        if start_index < len(source_text):
            current_string = InlineHelper.append_text(
                current_string, source_text[start_index:]
            )
            POGGER.debug("__cibp>current_string>$<", current_string)

        if current_string or not have_processed_once:
            POGGER.debug("__cibp>current_string>$<", current_string)
            POGGER.debug("__cibp>starting_whitespace>$<", starting_whitespace)
            if (
                is_setext
                and end_string is None
                and (inline_blocks and inline_blocks[-1].is_inline_hard_break)
            ):
                new_index, ex_ws = ParserHelper.extract_whitespace(current_string, 0)
                POGGER.debug("__cibp>new_index>$<", new_index)
                POGGER.debug("__cibp>b>$<", ex_ws)
                if new_index:
                    end_string = ex_ws + ParserHelper.whitespace_split_character
                    current_string = current_string[new_index:]
            POGGER.debug("__cibp>end_string>$<", end_string)
            inline_blocks.append(
                TextMarkdownToken(
                    current_string,
                    starting_whitespace,
                    end_whitespace=end_string,
                    line_number=line_number,
                    column_number=column_number,
                )
            )
        POGGER.debug(">>$<<", inline_blocks)

        return EmphasisHelper.resolve_inline_emphasis(inline_blocks, None)

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
https://github.github.com/gfm/#backslash-escapes
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_backslash_escapes_308():
    """
    Test case 308:  Any ASCII punctuation character may be backslash-escaped:
    """

    # Arrange
    source_markdown = """\\!\\\"\\#\\$\\%\\&\\'\\(\\)\\*\\+\\,\\-\\.\\/\\:\\;\\<\\=\\>\\?\\@\\[\\\\\\]\\^\\_\\`\\{\\|\\}\\~"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\\\b!\\\b\a\"\a&quot;\a\\\b#\\\b$\\\b%\\\b\a&\a&amp;\a\\\b'\\\b(\\\b)\\\b*\\\b+\\\b,\\\b-\\\b.\\\b/\\\b:\\\b;\\\b\a<\a&lt;\a\\\b=\\\b\a>\a&gt;\a\\\b?\\\b@\\\b[\\\b\\\\\b]\\\b^\\\b_\\\b`\\\b{\\\b|\\\b}\\\b~:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>!&quot;#$%&amp;'()*+,-./:;&lt;=&gt;?@[\\]^_`{|}~</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_309():
    """
    Test case 309:  Backslashes before other characters are treated as literal backslashes:
    """

    # Arrange
    source_markdown = """\\→\\A\\a\\ \\3\\φ\\«"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\\→\\A\\a\\ \\3\\φ\\«:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>\\→\\A\\a\\ \\3\\φ\\«</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_310():
    """
    Test case 310:  Escaped characters are treated as regular characters and do not have their usual Markdown meanings:
    """

    # Arrange
    source_markdown = """\\*not emphasized*
\\<br/> not a tag
\\[not a link](/foo)
\\`not code`
1\\. not a list
\\* not a list
\\# not a heading
\\[foo]: /url "not a reference"
\\&ouml; not a character entity"""
    expected_tokens = [
        "[para(1,1):\n\n\n\n\n\n\n\n]",
        "[text(1,1):\\\b*not emphasized:]",
        "[text(1,17):*:]",
        "[text(1,18):\n\\\b\a<\a&lt;\abr/\a>\a&gt;\a not a tag\n\\\b[not a link::\n\n]",
        "[text(3,13):]:]",
        """[text(3,14):(/foo)
\\\b`not code`
1\\\b. not a list
\\\b* not a list
\\\b# not a heading
\\\b[foo::\n\n\n\n\n]""",
        "[text(8,6):]:]",
        """[text(8,7):: /url \a\"\a&quot;\anot a reference\a\"\a&quot;\a
\\\b\a&\a&amp;\aouml; not a character entity::\n]""",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*not emphasized*
&lt;br/&gt; not a tag
[not a link](/foo)
`not code`
1. not a list
* not a list
# not a heading
[foo]: /url &quot;not a reference&quot;
&amp;ouml; not a character entity</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_311():
    """
    Test case 311:  If a backslash is itself escaped, the following character is not:
    """

    # Arrange
    source_markdown = """\\\\*emphasis*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\\\b\\:]",
        "[emphasis(1,3):1:*]",
        "[text(1,4):emphasis:]",
        "[end-emphasis(1,12)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>\\<em>emphasis</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_312():
    """
    Test case 312:  A backslash at the end of the line is a hard line break:
    """

    # Arrange
    source_markdown = """foo\\
bar"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):foo:]",
        "[hard-break(1,4):\\:\n]",
        "[text(2,1):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo<br />
bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_313():
    """
    Test case 313:  (part 1) Backslash escapes do not work in code blocks, code spans, autolinks, or raw HTML:
    """

    # Arrange
    source_markdown = """`` \\[\\` ``"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):\\[\\`:``: : ]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>\\[\\`</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_314():
    """
    Test case 314:  (part 2) Backslash escapes do not work in code blocks, code spans, autolinks, or raw HTML:
    """

    # Arrange
    source_markdown = """    \\[\\]"""
    expected_tokens = [
        "[icode-block(1,5):    :]",
        "[text(1,5):\\[\\]:]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<pre><code>\\[\\]
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_315():
    """
    Test case 315:  (part 3) Backslash escapes do not work in code blocks, code spans, autolinks, or raw HTML:
    """

    # Arrange
    source_markdown = """~~~
\\[\\]
~~~"""
    expected_tokens = [
        "[fcode-block(1,1):~:3::::::]",
        "[text(2,1):\\[\\]:]",
        "[end-fcode-block::3:False]",
    ]
    expected_gfm = """<pre><code>\\[\\]
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_316():
    """
    Test case 316:  (part 4) Backslash escapes do not work in code blocks, code spans, autolinks, or raw HTML:
    """

    # Arrange
    source_markdown = """<http://example.com?find=\\*>"""
    expected_tokens = [
        "[para(1,1):]",
        "[uri-autolink(1,1):http://example.com?find=\\*]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="http://example.com?find=%5C*">http://example.com?find=\\*</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_317():
    """
    Test case 317:  (part 5) Backslash escapes do not work in code blocks, code spans, autolinks, or raw HTML:
    """

    # Arrange
    source_markdown = """<a href="/bar\\/)">"""
    expected_tokens = [
        "[html-block(1,1)]",
        '[text(1,1):<a href="/bar\\/)">:]',
        "[end-html-block:::True]",
    ]
    expected_gfm = """<a href="/bar\\/)">"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_318():
    """
    Test case 318:  (part 1) But they work in all other contexts, including URLs and link titles, link references, and info strings in fenced code blocks:
    """

    # Arrange
    source_markdown = """[foo](/bar\\* "ti\\*tle")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/bar*:ti*tle:/bar\\*:ti\\*tle::foo:False:":: :]',
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/bar*" title="ti*tle">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_319():
    """
    Test case 319:  (part 2) But they work in all other contexts, including URLs and link titles, link references, and info strings in fenced code blocks:
    """

    # Arrange
    source_markdown = """[foo]

[foo]: /bar\\* "ti\\*tle"
"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/bar*:ti*tle::::foo:::::]",
        "[text(1,2):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        '[link-ref-def(3,1):True::foo:: :/bar*:/bar\\*: :ti*tle:"ti\\*tle":]',
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p><a href="/bar*" title="ti*tle">foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_320():
    """
    Test case 320:  (part 3) But they work in all other contexts, including URLs and link titles, link references, and info strings in fenced code blocks:
    """

    # Arrange
    source_markdown = """``` foo\\+bar
foo
```"""
    expected_tokens = [
        "[fcode-block(1,1):`:3:foo+bar:foo\\+bar:::: ]",
        "[text(2,1):foo:]",
        "[end-fcode-block::3:False]",
    ]
    expected_gfm = """<pre><code class="language-foo+bar">foo
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_320a():
    """
    Test case 320a:  (part 3) But they work in all other contexts, including URLs and link titles, link references, and info strings in fenced code blocks:
    """

    # Arrange
    source_markdown = """``` foo\\+\\bar
foo
```"""
    expected_tokens = [
        "[fcode-block(1,1):`:3:foo+\\bar:foo\\+\\bar:::: ]",
        "[text(2,1):foo:]",
        "[end-fcode-block::3:False]",
    ]
    expected_gfm = """<pre><code class="language-foo+\\bar">foo
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_320b():
    """
    Test case 320b:  (part 3) But they work in all other contexts, including URLs and link titles, link references, and info strings in fenced code blocks:
    """

    # Arrange
    source_markdown = """``` foo\\+bar\\
foo
```"""
    expected_tokens = [
        "[fcode-block(1,1):`:3:foo+bar\\:foo\\+bar\\:::: ]",
        "[text(2,1):foo:]",
        "[end-fcode-block::3:False]",
    ]
    expected_gfm = """<pre><code class="language-foo+bar\\">foo
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_backslash_escapes_320c():
    """
    Test case 320c:  (part 3) But they work in all other contexts, including URLs and link titles, link references, and info strings in fenced code blocks:
    """

    # Arrange
    source_markdown = """``` foo \\+bar\\
foo
```"""
    expected_tokens = [
        "[fcode-block(1,1):`:3:foo:: +bar\\: \\+bar\\:: ]",
        "[text(2,1):foo:]",
        "[end-fcode-block::3:False]",
    ]
    expected_gfm = """<pre><code class="language-foo">foo
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide for a leaf element that can be added to markdown parsing stream.
"""
from pymarkdown.markdown_token import MarkdownToken, MarkdownTokenClass
from pymarkdown.parser_helper import ParserHelper


class LeafMarkdownToken(MarkdownToken):
    """
    Class to provide for a leaf element that can be added to markdown parsing stream.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_name,
        extra_data,
        line_number=0,
        column_number=0,
        position_marker=None,
        extracted_whitespace=None,
        is_extension=False,
        requires_end_token=False,
        can_force_close=True,
    ):
        self.__extracted_whitespace = extracted_whitespace
        MarkdownToken.__init__(
            self,
            token_name,
            MarkdownTokenClass.LEAF_BLOCK,
            extra_data,
            line_number=line_number,
            column_number=column_number,
            position_marker=position_marker,
            is_extension=is_extension,
            requires_end_token=requires_end_token,
            can_force_close=can_force_close,
        )

    # pylint: enable=too-many-arguments

    @property
    def extracted_whitespace(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace


# pylint: disable=too-few-public-methods
class BlankLineMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the blank line element.
    """

    def __init__(self, extracted_whitespace, position_marker, column_delta=0):

        if position_marker:
            line_number, column_number = position_marker.line_number, (
                position_marker.index_number
                + position_marker.index_indent
                + 1
                - column_delta
            )
        else:
            line_number, column_number = 0, 0

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_blank_line,
            extracted_whitespace,
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=extracted_whitespace,
        )


# pylint: enable=too-few-public-methods


class ParagraphMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the paragraph element.
    """

    def __init__(self, extracted_whitespace, position_marker):
        self.__extracted_whitespace, self.__final_whitespace, self.rehydrate_index = (
            extracted_whitespace,
            "",
            0,
        )
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_paragraph,
            "",
            position_marker=position_marker,
            requires_end_token=True,
        )
        self.__compose_extra_data_field()

    @property
    def extracted_whitespace(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def final_whitespace(self):
        """
        Returns any final whitespace at the end of the paragraph that was removed.
        """
        return self.__final_whitespace

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        self._set_extra_data(
            f"{self.__extracted_whitespace}{MarkdownToken.extra_data_separator}{self.__final_whitespace}"
            if self.final_whitespace
            else self.__extracted_whitespace
        )

    def add_whitespace(self, whitespace_to_add):
        """
        Add extra whitespace to the end of the current paragraph.  Should only be
        used when combining text blocks in a paragraph.
        """

        self.__extracted_whitespace = (
            f"{self.__extracted_whitespace}{whitespace_to_add}"
        )
        self.__compose_extra_data_field()

    def set_final_whitespace(self, whitespace_to_set):
        """
        Set the final whitespace for the paragraph. That is any whitespace at the very
        end of the paragraph, removed to prevent hard lines at the end.
        """

        self.__final_whitespace = whitespace_to_set
        self.__compose_extra_data_field()


class ThematicBreakMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the thematic break element.
    """

    def __init__(
        self, start_character, extracted_whitespace, rest_of_line, position_marker
    ):
        self.__rest_of_line = rest_of_line
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_thematic_break,
            MarkdownToken.extra_data_separator.join(
                [start_character, extracted_whitespace, self.__rest_of_line]
            ),
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
        )

    @property
    def rest_of_line(self):
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__rest_of_line


class HtmlBlockMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the html block element.
    """

    def __init__(self, position_marker, extracted_whitespace):
        if position_marker:
            line_number, column_number = position_marker.line_number, (
                position_marker.index_number
                + position_marker.index_indent
                + 1
                - len(extracted_whitespace)
            )
        else:
            line_number, column_number = -1, -1

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_html_block,
            "",
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
        )


# pylint: disable=too-many-instance-attributes
class LinkReferenceDefinitionMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the link reference definition element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        did_add_definition,
        extracted_whitespace,
        link_name,
        link_value,
        link_debug,
        position_marker,
    ):
        self.__did_add_definition = did_add_definition
        self.__link_name = link_name

        if link_value:
            self.__link_destination, self.__link_title = link_value[0], link_value[1]
        else:
            self.__link_destination, self.__link_title = "", ""

        if link_debug:
            (
                self.__link_name_debug,
                self.__link_destination_whitespace,
                self.__link_destination_raw,
                self.__link_title_whitespace,
                self.__link_title_raw,
                self.__end_whitespace,
            ) = (
                "" if link_debug[0] == self.__link_name else link_debug[0],
                link_debug[1],
                "" if link_debug[2] == self.__link_destination else link_debug[2],
                link_debug[3],
                "" if link_debug[4] == self.__link_title else link_debug[4],
                link_debug[5],
            )
        else:
            (
                self.__link_name_debug,
                self.__link_destination_whitespace,
                self.__link_destination_raw,
                self.__link_title_whitespace,
                self.__link_title_raw,
                self.__end_whitespace,
            ) = ("", "", "", "", "", "")

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_link_reference_definition,
            MarkdownToken.extra_data_separator.join(
                [
                    str(self.did_add_definition),
                    extracted_whitespace,
                    self.__link_name,
                    self.__link_name_debug,
                    self.__link_destination_whitespace,
                    self.__link_destination,
                    self.__link_destination_raw,
                    self.__link_title_whitespace,
                    self.__link_title,
                    self.__link_title_raw,
                    self.__end_whitespace,
                ]
            ),
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
        )

    # pylint: enable=too-many-arguments
    @property
    def did_add_definition(self):
        """
        Returns an indication of whether the definition was actually added.
        """
        return self.__did_add_definition

    @property
    def end_whitespace(self):
        """
        Returns any whitespace that was extracted after the processing of this element occurred.
        """
        return self.__end_whitespace

    @property
    def link_name(self):
        """
        Returns the name of the link that was defined.
        """
        return self.__link_name

    @property
    def link_name_debug(self):
        """
        Returns the name of the link that was defined, in debug form.
        """
        return self.__link_name_debug

    @property
    def link_destination_whitespace(self):
        """
        Returns the whitespace that occurs before the link destination.
        """
        return self.__link_destination_whitespace

    @property
    def link_destination(self):
        """
        Returns the destination (URI) of the link that was defined.
        """
        return self.__link_destination

    @property
    def link_destination_raw(self):
        """
        Returns the destination (URI) of the link that was defined, in raw form.
        """
        return self.__link_destination_raw

    @property
    def link_title(self):
        """
        Returns the title of the link that was defined.
        """
        return self.__link_title

    @property
    def link_title_raw(self):
        """
        Returns the title of the link that was defined, in raw form.
        """
        return self.__link_title_raw

    @property
    def link_title_whitespace(self):
        """
        Returns the whitespace that occurs after the link title.
        """
        return self.__link_title_whitespace


# pylint: enable=too-many-instance-attributes


class AtxHeadingMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the atx heading element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        hash_count,
        remove_trailing_count,
        extracted_whitespace,
        position_marker,
    ):
        self.__hash_count, self.__remove_trailing_count = (
            hash_count,
            remove_trailing_count,
        )

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_atx_heading,
            "",
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
            can_force_close=False,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments

    @property
    def hash_count(self):
        """
        Returns the number of hash marks specified at the start of the line.
        """
        return self.__hash_count

    @property
    def remove_trailing_count(self):
        """
        Returns the number of hash marks specified at the end of the line.
        """
        return self.__remove_trailing_count

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [
                    str(self.__hash_count),
                    str(self.__remove_trailing_count),
                    self.extracted_whitespace,
                ]
            )
        )


# pylint: disable=too-many-instance-attributes
class SetextHeadingMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the setext heading element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        heading_character,
        heading_character_count,
        extracted_whitespace,
        position_marker,
        para_token,
    ):
        (
            self.__heading_character,
            self.__heading_character_count,
            self.__final_whitespace,
            self.__original_line_number,
            self.__original_column_number,
        ) = (
            heading_character,
            heading_character_count,
            "",
            para_token.line_number if para_token else -1,
            para_token.column_number if para_token else -1,
        )

        if self.__heading_character == "=":
            self.__hash_count = 1
        elif self.__heading_character == "-":
            self.__hash_count = 2
        else:
            self.__hash_count = -1

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_setext_heading,
            "",
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
            can_force_close=False,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments

    @property
    def final_whitespace(self):
        """
        Returns any final whitespace at the end of the heading that was removed.
        """
        return self.__final_whitespace

    @property
    def heading_character(self):
        """
        Returns the character associated with the heading start.
        """
        return self.__heading_character

    @property
    def hash_count(self):
        """
        Returns the count in equivalence of "Atx Hash" counts.
        """
        return self.__hash_count

    @property
    def heading_character_count(self):
        """
        Returns the count of characters associated with the heading start.
        """
        return self.__heading_character_count

    @property
    def original_line_number(self):
        """
        Returns the line number where this element actually started.
        """
        return self.__original_line_number

    @property
    def original_column_number(self):
        """
        Returns the column number where this element actually started.
        """
        return self.__original_column_number

    def set_final_whitespace(self, whitespace_to_set):
        """
        Set the final whitespace for the paragraph. That is any whitespace at the very
        end of the paragraph, removed to prevent hard lines at the end.
        """

        self.__final_whitespace = whitespace_to_set
        self.__compose_extra_data_field()

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        original_location = (
            f"({str(self.original_line_number)},{str(self.original_column_number)})"
        )
        field_parts = [
            self.__heading_character,
            str(self.__heading_character_count),
            self.extracted_whitespace,
            original_location,
        ]
        if self.final_whitespace:
            field_parts.append(self.final_whitespace)
        self._set_extra_data(MarkdownToken.extra_data_separator.join(field_parts))


# pylint: enable=too-many-instance-attributes


class IndentedCodeBlockMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the indented code block element.
    """

    def __init__(self, extracted_whitespace, line_number, column_number):
        self.__indented_whitespace = ""
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_indented_code_block,
            extracted_whitespace,
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
        )
        self.__compose_extra_data_field()

    @property
    def indented_whitespace(self):
        """
        Returns any indented whitespace that comes before the text.
        """
        return self.__indented_whitespace

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [self.extracted_whitespace, self.indented_whitespace]
            )
        )

    def add_indented_whitespace(self, indented_whitespace):
        """
        Add the indented whitespace that comes before the text.
        """
        self.__indented_whitespace = f"{self.__indented_whitespace}{ParserHelper.newline_character}{indented_whitespace}"
        self.__compose_extra_data_field()


# pylint: disable=too-many-instance-attributes
class FencedCodeBlockMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the fenced code block element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        fence_character,
        fence_count,
        extracted_text,
        pre_extracted_text,
        text_after_extracted_text,
        pre_text_after_extracted_text,
        extracted_whitespace,
        extracted_whitespace_before_info_string,
        position_marker,
    ):
        (
            self.__extracted_text,
            self.__pre_extracted_text,
            self.__extracted_whitespace_before_info_string,
            self.__text_after_extracted_text,
            self.__pre_text_after_extracted_text,
            self.__fence_character,
            self.__fence_count,
        ) = (
            extracted_text,
            pre_extracted_text,
            extracted_whitespace_before_info_string,
            text_after_extracted_text,
            pre_text_after_extracted_text,
            fence_character,
            fence_count,
        )
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_fenced_code_block,
            "",
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments

    @property
    def fence_character(self):
        """
        Returns the character used for the fence.
        """
        return self.__fence_character

    @property
    def fence_count(self):
        """
        Returns the number of fence characters used for the fence.
        """
        return self.__fence_count

    @property
    def extracted_text(self):
        """
        Returns the text extracted from the info string.
        """
        return self.__extracted_text

    @property
    def pre_extracted_text(self):
        """
        Returns the text extracted from the info string.
        """
        return self.__pre_extracted_text

    @property
    def text_after_extracted_text(self):
        """
        Returns the text extracted after the info string.
        """
        return self.__text_after_extracted_text

    @property
    def pre_text_after_extracted_text(self):
        """
        Returns the text extracted after after the info string.
        """
        return self.__pre_text_after_extracted_text

    @property
    def extracted_whitespace_before_info_string(self):
        """
        Returns any whitespace that was extracted before the info string was processed.
        """
        return self.__extracted_whitespace_before_info_string

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [
                    self.__fence_character,
                    str(self.__fence_count),
                    self.__extracted_text,
                    self.__pre_extracted_text,
                    self.__text_after_extracted_text,
                    self.__pre_text_after_extracted_text,
                    self.extracted_whitespace,
                    self.__extracted_whitespace_before_info_string,
                ]
            )
        )


# pylint: enable=too-many-instance-attributes
